"""Verify that the new distances functions match the original ones."""

import math
from collections.abc import Callable

import pytest
import torch
from hypothesis import given
from hypothesis import strategies as st
from torch import Tensor
from torch.testing import assert_close, make_tensor

from fastabx.distance import DistanceName, distance_function

BATCH, SEQ, DIM = st.integers(1, 20), st.integers(1, 10), st.integers(1, 1024)
LOW, HIGH_MINUS_LOW = st.floats(-100, 100), st.floats(0.1, 100)
MIN_POSITIVE = 0.1


def kl_distance(a1: Tensor, a2: Tensor, epsilon: float = 1e-6) -> Tensor:
    """KL distance. You might want to use kl_symmetric_distance in most cases."""
    n1, s1, d = a1.size()
    n2, s2, d = a2.size()
    div = (a1.view(n1, 1, s1, 1, d) + epsilon) / (a2.view(1, n2, 1, s2, d) + epsilon)
    prod = (a1.view(n1, 1, s1, 1, d)) * div.log()
    return prod.sum(dim=4)


def kl_symmetric_distance(a1: Tensor, a2: Tensor, epsilon: float = 1e-6) -> Tensor:
    """KL symmetric distance. The two tensors must correspond to probability distributions."""
    n1, s1, d = a1.size()
    n2, s2, d = a2.size()
    div1 = (a1.view(n1, 1, s1, 1, d) + epsilon) / (a2.view(1, n2, 1, s2, d) + epsilon)
    div2 = (a2.view(1, n2, 1, s2, d) + epsilon) / (a1.view(n1, 1, s1, 1, d) + epsilon)
    prod1 = (a1.view(n1, 1, s1, 1, d)) * div1.log()
    prod2 = (a2.view(1, n2, 1, s2, d)) * div2.log()
    return (0.5 * prod1 + 0.5 * prod2).sum(dim=4)


def cosine_distance(a1: Tensor, a2: Tensor) -> Tensor:
    """Angular distance (default). WARNING: a1 and a2 must be normalized."""
    n1, s1, d = a1.size()
    n2, s2, d = a2.size()
    prod = (a1.view(n1, 1, s1, 1, d)) * (a2.view(1, n2, 1, s2, d))
    return torch.clamp(prod.sum(dim=4), -1, 1).acos() / math.pi


def euclidean_distance(a1: Tensor, a2: Tensor) -> Tensor:
    """Euclidean distance."""
    n1, s1, d = a1.size()
    n2, s2, d = a2.size()
    diff = a1.view(n1, 1, s1, 1, d) - a2.view(1, n2, 1, s2, d)
    return torch.sqrt((diff**2).sum(dim=4))


OLD_DISTANCE_FN: dict[str, Callable[[Tensor, Tensor], Tensor]] = {
    "kl": kl_distance,
    "kl_symmetric": kl_symmetric_distance,
    "euclidean": euclidean_distance,
    "cosine": cosine_distance,
}


@pytest.mark.parametrize("name", list(OLD_DISTANCE_FN.keys()))
@given(n1=BATCH, n2=BATCH, s1=SEQ, s2=SEQ, d=DIM, low=LOW, high_minus_low=HIGH_MINUS_LOW)
def test_distance_new_implementation(
    name: DistanceName,
    n1: int,
    n2: int,
    s1: int,
    s2: int,
    d: int,
    low: float,
    high_minus_low: float,
) -> None:
    """Compare the new distance implementation to the old one."""
    a = make_tensor((n1, s1, d), dtype=torch.float32, low=low, high=high_minus_low + low, device="cpu")
    b = make_tensor((n2, s2, d), dtype=torch.float32, low=low, high=high_minus_low + low, device="cpu")
    if name.startswith("kl"):
        a, b = torch.clamp(a, min=MIN_POSITIVE), torch.clamp(b, min=MIN_POSITIVE)
    distance_old = OLD_DISTANCE_FN[name]
    distance_new = distance_function(name)
    assert_close(distance_new(a, b), distance_old(a, b))



"""Additional constraints for building cells."""

import functools
import operator
from collections.abc import Generator, Iterable

import polars as pl
import polars.selectors as cs
import torch
from polars._typing import IntoExprColumn
from tqdm import tqdm

from fastabx.cell import Cell
from fastabx.distance import Distance, abx_on_cell
from fastabx.task import Task
from fastabx.utils import MIN_CELLS_FOR_TQDM

type Constraints = Iterable[IntoExprColumn]


def constraints_all_different(*columns: str) -> Constraints:
    """Return :py:type:`.Constraints` that ensure that each specified column has different values for A, B and X."""
    return [
        pl.col(f"{c}_a").ne(pl.col(f"{c}_x"))
        & pl.col(f"{c}_a").ne(pl.col(f"{c}_b"))
        & pl.col(f"{c}_x").ne(pl.col(f"{c}_b"))
        for c in columns
    ]


class NoConstraintsError(ValueError):
    """Invalid constraints."""

    def __init__(self) -> None:
        super().__init__("No valid column provided in the constraints")


def apply_constraints(
    cells: pl.DataFrame,
    labels: pl.DataFrame,
    constraints: Constraints,
    *,
    is_symmetric: bool,
) -> pl.DataFrame:
    """Apply constraints to the cells DataFrame."""
    columns_to_retrieve = {
        name.removesuffix("_x").removesuffix("_a").removesuffix("_b")
        for constraint in constraints
        for name in constraint.meta.root_names()
    }
    if not columns_to_retrieve or not columns_to_retrieve.issubset(labels.columns):
        raise NoConstraintsError
    if is_symmetric:
        constraints = [*constraints, pl.col("index_a") != pl.col("index_x")]
    labels_lazy = labels.lazy().select(*columns_to_retrieve).with_row_index()
    cells_lazy = cells.lazy()
    is_valid = (
        cells_lazy.explode("index_x")
        .explode("index_a")
        .explode("index_b")
        .join(labels_lazy.rename({c: f"{c}_x" for c in (columns_to_retrieve | {"index"})}), on="index_x")
        .join(labels_lazy.rename({c: f"{c}_a" for c in (columns_to_retrieve | {"index"})}), on="index_a")
        .join(labels_lazy.rename({c: f"{c}_b" for c in (columns_to_retrieve | {"index"})}), on="index_b")
        .with_columns(is_valid=functools.reduce(operator.and_, constraints))
        .select(cs.exclude([f"{c}_{s}" for c in columns_to_retrieve for s in ("a", "b", "x")]))
        .group_by(cs.exclude(cs.starts_with("index_") | pl.col("is_valid")), maintain_order=True)
        .agg("is_valid")
        .select("is_valid")
    )
    return pl.concat((cells_lazy, is_valid), how="horizontal").collect()


def constrained_cell_generator(
    task: Task, constraints: Constraints
) -> Generator[tuple[Cell, torch.Tensor], None, None]:
    """Generate cells with constraints applied, yielding (Cell, mask) tuples."""
    is_symmetric, device = not bool(task.across), task.dataset.accessor.device
    cells = apply_constraints(task.cells, task.dataset.labels, constraints, is_symmetric=is_symmetric)
    columns = ["header", "description", "index_a", "index_b", "index_x", "is_valid"]
    for header, description, index_a, index_b, index_x, is_valid in cells[columns].iter_rows():
        a = task.dataset.accessor.batched(index_a)
        b = task.dataset.accessor.batched(index_b)
        x = task.dataset.accessor.batched(index_x)
        mask = torch.tensor(is_valid, device=device).view((len(x.sizes), len(a.sizes), len(b.sizes)))
        yield Cell(a=a, b=b, x=x, header=header, description=description, is_symmetric=is_symmetric), mask


def score_task_with_constraints(
    task: Task, distance: Distance, constraints: Constraints
) -> tuple[list[float], list[int]]:
    """Score each cell of a :py:class:`.Task` with additional constraints."""
    scores, sizes = [], []
    for cell, mask in tqdm(
        constrained_cell_generator(task, constraints),
        "Scoring each cell with constraints",
        total=len(task),
        disable=len(task) < MIN_CELLS_FOR_TQDM,
    ):
        if not mask.any():
            scores.append(None)
            sizes.append(None)
        else:
            scores.append(abx_on_cell(cell, distance, mask=mask).item())
            sizes.append(mask.sum())
    return scores, sizes



"""Distance computation."""

import math
from collections.abc import Callable
from typing import Literal, get_args

import torch
from torch import Tensor

from fastabx.cell import Cell
from fastabx.dtw import dtw_batch

type Distance = Callable[[Tensor, Tensor], Tensor]
type DistanceName = Literal["euclidean", "cosine", "angular", "kl", "kl_symmetric", "identical", "null"]


def available_distances() -> tuple[str, ...]:
    """Names of the available distances."""
    return get_args(DistanceName.__value__)


def distance_function(name: DistanceName) -> Distance:
    """Return the corresponding distance function."""
    match name:
        case "euclidean":
            return euclidean_distance
        case "cosine" | "angular":
            return angular_distance
        case "kl":
            return kl_distance
        case "kl_symmetric":
            return kl_symmetric_distance
        case "identical":
            return identical_distance
        case "null":
            return null_distance
        case _:
            raise ValueError(name)


def null_distance(a1: Tensor, a2: Tensor) -> Tensor:
    """Null distance, useful for debugging."""
    n1, s1, _ = a1.size()
    n2, s2, _ = a2.size()
    return torch.zeros(n1, n2, s1, s2, device=a1.device)


def kl_distance(a1: Tensor, a2: Tensor, epsilon: float = 1e-6) -> Tensor:
    """KL distance. You might want to use kl_symmetric_distance in most cases."""
    n1, s1, d = a1.size()
    n2, s2, d = a2.size()
    a1_view = a1.view(n1 * s1, 1, d)
    a2_view = a2.view(1, n2 * s2, d)
    div = (a1_view + epsilon) / (a2_view + epsilon)
    return torch.sum(a1_view * div.log(), dim=2).view(n1, s1, n2, s2).transpose(1, 2)


def kl_symmetric_distance(a1: Tensor, a2: Tensor, epsilon: float = 1e-6) -> Tensor:
    """KL symmetric distance. The two tensors must correspond to probability distributions."""
    n1, s1, d = a1.size()
    n2, s2, d = a2.size()
    a1_view = a1.view(n1 * s1, 1, d)
    a2_view = a2.view(1, n2 * s2, d)
    div1 = (a1_view + epsilon) / (a2_view + epsilon)
    div2 = (a2_view + epsilon) / (a1_view + epsilon)
    out1 = torch.sum(a1_view * div1.log(), dim=2)
    out2 = torch.sum(a2_view * div2.log(), dim=2)
    return (0.5 * out1 + 0.5 * out2).view(n1, s1, n2, s2).transpose(1, 2)


def angular_distance(a1: Tensor, a2: Tensor) -> Tensor:
    """Angular distance (default). WARNING: a1 and a2 must be normalized."""
    n1, s1, d = a1.size()
    n2, s2, d = a2.size()
    dot_prods = torch.mm(a1.view(n1 * s1, d), a2.view(n2 * s2, d).T).view(n1, s1, n2, s2).transpose(1, 2)
    return torch.clamp(dot_prods, -1, 1).acos() / math.pi


def euclidean_distance(a1: Tensor, a2: Tensor) -> Tensor:
    """Euclidean distance."""
    n1, s1, d = a1.size()
    n2, s2, d = a2.size()
    dist = torch.cdist(a1.view(n1 * s1, d), a2.view(n2 * s2, d), compute_mode="donot_use_mm_for_euclid_dist")
    return dist.view(n1, s1, n2, s2).transpose(1, 2)


def identical_distance(a1: Tensor, a2: Tensor) -> Tensor:
    """0/1 distance. Useful for computing the ABX on discrete speech units."""
    n1, s1, _ = a1.size()
    n2, s2, _ = a2.size()
    return (a1.view(n1, 1, s1, 1) != a2.view(1, n2, 1, s2)).float()


def distance_on_cell(cell: Cell, distance: Distance) -> tuple[torch.Tensor, torch.Tensor]:
    """Compute the distance matrices between all A and X, and all B and X in the ``cell``, for a given ``distance``."""
    (a, sa), (b, sb), (x, sx) = (cell.a.data, cell.a.sizes), (cell.b.data, cell.b.sizes), (cell.x.data, cell.x.sizes)
    if cell.use_dtw:
        dxa = dtw_batch(distance(x, a), sx, sa, symmetric=cell.is_symmetric)
        dxb = dtw_batch(distance(x, b), sx, sb, symmetric=False)
    else:
        dxa, dxb = distance(x, a).squeeze(2, 3), distance(x, b).squeeze(2, 3)
    return dxa, dxb


def abx_on_cell(cell: Cell, distance: Distance, *, mask: torch.Tensor | None = None) -> torch.Tensor:
    """Compute the ABX of a ``cell`` using the given ``distance``."""
    dxa, dxb = distance_on_cell(cell, distance)
    if cell.is_symmetric:
        dxa.fill_diagonal_(float("inf"))
    nx, na = dxa.size()
    nx, nb = dxb.size()
    dxb = dxb.view(nx, 1, nb).expand(nx, na, nb)
    dxa = dxa.view(nx, na, 1).expand(nx, na, nb)
    if mask is None:
        sc = (dxa < dxb).sum() + 0.5 * (dxa == dxb).sum()
        sc /= len(cell)
    else:
        sc = (dxa < dxb)[mask].sum() + 0.5 * (dxa == dxb)[mask].sum()
        sc /= mask.sum()
    return 1 - sc



"""Task module. The Task class builds all the cells for the 'by', 'on' and 'across' conditions."""

from dataclasses import dataclass

import polars as pl
import polars.selectors as cs

from fastabx.dataset import Batch
from fastabx.verify import verify_cell

MIN_A_LEN = 2  # Minimum length of A in the ABX task.


@dataclass(frozen=True)
class Cell:
    """Individual cell of the ABX task."""

    a: Batch
    b: Batch
    x: Batch
    header: str
    description: str
    is_symmetric: bool

    def __post_init__(self) -> None:
        verify_cell((self.a.data, self.a.sizes), (self.b.data, self.b.sizes), (self.x.data, self.x.sizes))

    @property
    def num_triplets(self) -> int:
        """Get the number of triplets in the cell."""
        nx = self.a.data.size(0) - 1 if self.is_symmetric else self.x.data.size(0)
        return self.a.data.size(0) * self.b.data.size(0) * nx

    @property
    def use_dtw(self) -> bool:
        """Whether or not to use the DTW when computing the distances for this cell."""
        return not (1 == self.a.data.size(1) == self.b.data.size(1) == self.x.data.size(1))

    def __len__(self) -> int:
        return self.num_triplets

    def __repr__(self) -> str:
        return "Cell(\n\t" + ")\n\t".join(self.description.split("), ")) + "\n)"


def cells_on_by(df: pl.LazyFrame, on: str, by: list[str]) -> pl.LazyFrame:
    """Generate the cells for the ABX task given 'on' and 'by' conditions.

    The 'by' condition is optional, the function still works if 'by' is an empty list.
    """
    df = (
        df.select([on, *by])
        .with_row_index()
        .group_by([on, *by], maintain_order=True)
        .agg(pl.col("index"))
        .with_columns(len=pl.col("index").list.len())
        .with_row_index(name="__lookup")
    )
    lookup = df.select("__lookup", "index")
    cells = df.select(cs.exclude("index"))
    cells = (
        cells.join(cells, on=by or None, suffix="_b", how="inner" if by else "cross")  # A_by == B_by
        .filter(pl.col(on) != pl.col(f"{on}_b"), pl.col("len") >= MIN_A_LEN)  # A_on != B_on,  A_len >= MIN_A_LEN
        .select(cs.exclude(f"{on}_x", "len", "len_b"))
    )
    return (
        cells.join(lookup.rename({"index": "index_a"}), on="__lookup", how="left")
        .join(lookup.rename({"__lookup": "__lookup_b", "index": "index_b"}), on="__lookup_b", how="left")
        .select(cs.exclude("__lookup", "__lookup_b"))
        .with_columns(index_x=pl.col("index_a"))
    )


class NoAcrossError(ValueError):
    """To raise if you expect 'across' conditions but received none."""

    def __init__(self) -> None:
        super().__init__("across must be non-empty")


def cells_on_by_across(df: pl.LazyFrame, on: str, by: list[str], across: list[str]) -> pl.LazyFrame:
    """Generate the cells for the ABX task given 'on', 'by' and 'across' conditions."""
    if not across:
        raise NoAcrossError
    df = (
        df.select([on, *by, *across])
        .with_row_index()
        .group_by([on, *by, *across], maintain_order=True)
        .agg(pl.col("index"))
        .with_row_index(name="__lookup")
    )
    lookup = df.select("__lookup", "index")
    cells = df.select(cs.exclude("index"))
    cells = (
        cells.join(cells, on=by + across, suffix="_b", how="inner")  # A_{by, across} == B_{by, across}
        .filter(pl.col(on) != pl.col(f"{on}_b"))  # A_on != B_on
        .join(cells, on=by or None, suffix="_x", how="inner" if by else "cross")  # A_by, B_by == X_by
        .filter(
            pl.col(on) == pl.col(f"{on}_x"),  # A_on == X_on
            pl.col(f"{on}_b") != pl.col(f"{on}_x"),  # B_on != X_on
            *[pl.col(c) != pl.col(f"{c}_x") for c in across],  # A_across, B_across != X_across
        )
        .select(cs.exclude(f"{on}_x"))
    )
    return (
        cells.join(lookup.rename({"index": "index_a"}), on="__lookup", how="left")
        .join(lookup.rename({"__lookup": "__lookup_b", "index": "index_b"}), on="__lookup_b", how="left")
        .join(lookup.rename({"__lookup": "__lookup_x", "index": "index_x"}), on="__lookup_x", how="left")
        .select(cs.exclude("__lookup", "__lookup_b", "__lookup_x"))
    )


def cell_description(on: str, by: list[str], across: list[str]) -> pl.Expr:
    """Long description of a cell."""
    cell = f"ON({on}_ax = " + pl.col(on).cast(pl.String) + f", {on}_b = " + pl.col(f"{on}_b").cast(pl.String) + ")"
    if by:
        cell += ", " + pl.concat_str([f"BY({c}_abx = " + pl.col(c).cast(pl.String) + ")" for c in by], separator=", ")
    if across:
        exprs = [
            f"{c}_ab = " + pl.col(c).cast(pl.String) + f", {c}_x = " + pl.col(f"{c}_x").cast(pl.String) for c in across
        ]
        cell += ", " + pl.format("ACROSS({})", pl.concat_str(exprs, separator=", "))
    return cell


def cell_header(on: str, by: list[str], across: list[str]) -> pl.Expr:
    """Header of a cell."""
    return pl.concat_str([on, f"{on}_b"] + by + across + [f"{c}_x" for c in across], separator="-")



# ruff: noqa: E402
"""Full ABX."""

from fastabx.utils import load_dtw_extension

load_dtw_extension()

from fastabx.dataset import Dataset
from fastabx.pooling import pooling
from fastabx.score import Score
from fastabx.subsample import Subsampler
from fastabx.task import Task
from fastabx.zerospeech import zerospeech_abx

__all__ = ["Dataset", "Score", "Subsampler", "Task", "pooling", "zerospeech_abx"]



"""Entry point for the ZeroSpeech ABX evaluation."""

import argparse
from argparse import ArgumentDefaultsHelpFormatter

from fastabx.distance import available_distances
from fastabx.zerospeech import zerospeech_abx


def main() -> None:
    """ZeroSpeech ABX evaluation."""
    parser = argparse.ArgumentParser(
        description="ZeroSpeech ABX",
        allow_abbrev=False,
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("item", help="Path to the item file")
    parser.add_argument("features", help="Path to the features directory")
    parser.add_argument(
        "--max-size-group",
        type=int,
        required=True,
        help="Maximum number of A, B, or X in a cell. Set to 10 in the original ZeroSpeech ABX. "
        "Disabled if negative value.",
    )
    parser.add_argument(
        "--max-x-across",
        type=int,
        help="With 'across', maximum number of X given (A, B). Set to 5 in the original ZeroSpeech ABX. "
        "Disabled if negative value.",
    )
    parser.add_argument("--frequency", type=int, default=50, help="Feature frequency (in Hz)")
    parser.add_argument("--speaker", choices=["within", "across"], default="within", help="Speaker mode")
    parser.add_argument("--context", choices=["within", "any"], default="within", help="Context mode")
    parser.add_argument("--distance", choices=available_distances(), default="angular", help="Distance")
    parser.add_argument("--seed", type=int, default=0, help="Random seed")
    args = parser.parse_args()

    if args.max_x_across is None and args.speaker == "across":
        parser.error("--max-x-across is required when using 'across' speaker mode")
    score = zerospeech_abx(
        args.item,
        args.features,
        max_size_group=args.max_size_group if args.max_size_group >= 0 else None,
        max_x_across=args.max_x_across if args.max_x_across is not None and args.max_x_across >= 0 else None,
        speaker=args.speaker,
        context=args.context,
        distance=args.distance,
        frequency=args.frequency,
        seed=args.seed,
    )
    print(f"ABX error rate: {score:.3%}")  # noqa: T201


if __name__ == "__main__":
    main()



"""ZeroSpeech ABX evaluation. Reproduces ZeroSpeech 2021."""

from collections.abc import Callable
from pathlib import Path
from typing import Literal

import torch

from fastabx.dataset import Dataset
from fastabx.distance import DistanceName
from fastabx.score import Score
from fastabx.subsample import Subsampler
from fastabx.task import Task


class InvalidSpeakerOrContextError(ValueError):
    """The speaker or context conditions are not set correctly."""


def zerospeech_abx(
    item: str | Path,
    root: str | Path,
    *,
    max_size_group: int | None,
    max_x_across: int | None,
    speaker: Literal["within", "across"] = "within",
    context: Literal["within", "any"] = "within",
    distance: DistanceName = "angular",
    frequency: int = 50,
    feature_maker: Callable[[str | Path], torch.Tensor] = torch.load,
    extension: str = ".pt",
    seed: int = 0,
) -> float:
    """Compute the ABX similarly to the ZeroSpeech 2021 challenge.

    On triphone or phoneme, described by an item file.
    Within or across speaker, and within context or ignoring context.

    :param item: the item file.
    :param root: the root directory containing either the features or the audio files.
    :param max_size_group: maximum number of instances of A, B, or X in each :py:class:`.Cell`.
        Passed to the :py:class:`.Subsampler` of the :py:class:`.Task`. Set to 10 in the original ZeroSpeech ABX code.
        Disabled if set to ``None``.
    :param max_x_across: in the "across" speaker mode, maximum number of X considered for given values of A and B.
        Passed to the :py:class:`.Subsampler` of the :py:class:`.Task`.
        Set to 5 in the original ZeroSpeech ABX code. Disabled if set to ``None``.
    :param speaker: the speaker mode, either "within" or "across". Defaults to "within".
    :param context: the context mode, either "within" or "any". Always use "within" with representations of triphones.
        Defaults to "within".
    :param distance: the distance metric, "angular" (same as "cosine"), "euclidean", "kl_symmetric" or "identical".
        Defaults to "angular".
    :param frequency: the feature frequency of the features / the output of the feature maker, in Hz.
        Defaults to 50 Hz.
    :param feature_maker: the feature maker. Defaults to just loading the file with ``torch.load``.
    :param extension: the filename extension of the files to process in ``root``, default is ".pt".
    :param seed: the random seed for the subsampling, default is 0.
    """
    dataset = Dataset.from_item(item, root, frequency, feature_maker=feature_maker, extension=extension)
    by: list[str] | None
    across: list[str] | None
    match (speaker, context):
        case ("within", "within"):
            by, across = ["prev-phone", "next-phone", "speaker"], None
        case ("within", "any"):
            by, across = ["speaker"], None
        case ("across", "within"):
            by, across = ["prev-phone", "next-phone"], ["speaker"]
        case ("across", "any"):
            by, across = None, ["speaker"]
        case _:
            raise InvalidSpeakerOrContextError
    subsampler = Subsampler(max_size_group, max_x_across, seed)
    task = Task(dataset, on="#phone", by=by, across=across, subsampler=subsampler)
    levels = ([("next-phone", "prev-phone")] if context == "within" else []) + ["speaker"]
    return Score(task, distance).collapse(levels=levels)



"""Data utilities."""

import abc
import math
from collections.abc import Callable, Collection, Iterable, Iterator, Mapping, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Self

import numpy as np
import numpy.typing as npt
import polars as pl
import polars.selectors as cs
import torch
from polars.interchange.protocol import SupportsInterchange
from torch.nn.utils.rnn import pad_sequence
from tqdm import tqdm

from fastabx.utils import with_librilight_bug
from fastabx.verify import verify_empty_datapoints

type ArrayLike = npt.ArrayLike


@dataclass(frozen=True)
class Batch:
    """Batch of padded data."""

    data: torch.Tensor
    sizes: torch.Tensor

    def __repr__(self) -> str:
        return f"Batch(data=Tensor(shape={self.data.shape}, dtype={self.data.dtype}), sizes={self.sizes})"


class DataAccessor(abc.ABC):
    """Abstract class for data accessors.

    A data accessor is a way to access a torch.Tensor given an index.
    """

    @abc.abstractmethod
    def __getitem__(self, i: int) -> torch.Tensor:
        pass

    @abc.abstractmethod
    def __len__(self) -> int:
        pass

    @abc.abstractmethod
    def __iter__(self) -> Iterator[torch.Tensor]:
        pass

    @abc.abstractmethod
    def batched(self, indices: Iterator[int]) -> Batch:
        """Get the padded data and the original sizes of the data from a list of indices."""


class InMemoryAccessor(DataAccessor):
    """Data accessor where everything is in memory."""

    def __init__(self, indices: dict[int, tuple[int, int]], data: torch.Tensor) -> None:
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.indices = indices
        verify_empty_datapoints(self.indices)
        self.data = data.to(self.device)

    def __repr__(self) -> str:
        return f"InMemoryAccessor(data of shape {tuple(self.data.shape)}, with {len(self)} items)"

    def __getitem__(self, i: int) -> torch.Tensor:
        if i not in self.indices:
            raise IndexError
        start, end = self.indices[i]
        return self.data[start:end]

    def __len__(self) -> int:
        return len(self.indices)

    def __iter__(self) -> Iterator[torch.Tensor]:
        for i in self.indices:
            yield self[i]

    def batched(self, indices: Iterator[int]) -> Batch:
        """Get the padded data and the original sizes of the data from a list of indices."""
        sizes, data = [], []
        for i in indices:
            this_data = self[i]
            sizes.append(this_data.size(0))
            data.append(this_data)
        return Batch(pad_sequence(data, batch_first=True), torch.tensor(sizes, dtype=torch.int64, device=self.device))


def find_all_files(root: str | Path, extension: str) -> dict[str, Path]:
    """Recursively find all files with the given `extension` in `root`."""
    return dict(sorted((p.stem, p) for p in Path(root).rglob(f"*{extension}")))


def normalize_with_singularity(x: torch.Tensor, eps: float = 1e-12) -> torch.Tensor:
    """Normalize the given vector across the third dimension.

    Extend all vectors by eps to put the null vector at the maximal
    angular distance from any non-null vector.
    """
    norm = torch.norm(x, dim=1, keepdim=True)
    zero_vals = norm == 0
    x = torch.where(zero_vals, 1 / math.sqrt(x.size(1)), x / norm)
    border = torch.full((x.size(0), 1), eps, dtype=x.dtype, device=x.device)
    border = torch.where(zero_vals, -2 * eps, border)
    return torch.cat([x, border], dim=1)


class InvalidItemFileError(Exception):
    """The item file does not have the correct columns."""


def read_item(item: str | Path) -> pl.DataFrame:
    """Read an item file."""
    schema = {
        "#file": pl.String,
        "onset": pl.String,
        "offset": pl.String,
        "#phone": pl.String,
        "prev-phone": pl.String,
        "next-phone": pl.String,
        "speaker": pl.String,
    }
    try:
        return pl.read_csv(item, separator=" ", schema=schema).with_columns(
            pl.col("onset").str.to_decimal(), pl.col("offset").str.to_decimal()
        )
    except pl.exceptions.ComputeError as error:
        raise InvalidItemFileError from error


def item_frontiers(frequency: float) -> tuple[pl.Expr, pl.Expr, pl.Expr, pl.Expr]:
    """Frontiers [start, end[ in the input features and in the concatenated ones."""
    # Remove the cast to float and -0.5 once RoundMode "half_ceil" and "half_floor" are added to polars round method
    # (or "half_away_from_zero" and "half_to_zero")
    # See: https://github.com/pola-rs/polars/issues/21800
    start = (pl.col("onset") * frequency - 0.5).cast(pl.Float64).ceil().cast(pl.Int64).alias("start")
    end = (pl.col("offset") * frequency - 0.5).cast(pl.Float64).floor().cast(pl.Int64).alias("end")
    if not with_librilight_bug():
        end += 1
    length = (end - start).alias("length")
    right = length.cum_sum().alias("right")
    left = length.cum_sum().shift(1).fill_null(0).alias("left")
    return start, end, left, right


class FeaturesSizeError(ValueError):
    """To raise if the features size is not correct."""

    def __init__(self, fileid: str, start: int, end: int, actual: int) -> None:
        super().__init__(
            f"Input features length is not correct for file {fileid}. It has a length {actual}, "
            f"but we are slicing between [{start}, {end}[.\n"
            f"The most common reason for this is that there is one frame missing in the features, because "
            f"of how the convolutional layers are defined in your model and because the phoneme under consideration "
            f"is at the very end of the file. You can either add padding to the convolutions, or add a bit of silence "
            f"at the end of the audio file."
        )


class EmptyFeaturesError(ValueError):
    """Raised when empty features are found when building the dataset."""

    def __init__(self, df: pl.DataFrame) -> None:
        super().__init__(
            f"{len(df)} empty entries found. These entries are shorter than a single unit at the given frequency. "
            f"First, check that the given frequency is correct. Then, if you intend to compute ABX on units this large"
            f", you must first remove these entries from your item file. "
            f"Refer to https://docs.cognitive-ml.fr/fastabx/advanced/slicing.html for details on how features are "
            f"sliced. The empty entries are: \n{df}"
        )


def load_data_from_item[T](
    mapping: dict[str, T],
    labels: pl.DataFrame,
    frequency: int,
    feature_maker: Callable[[T], torch.Tensor],
) -> tuple[dict[int, tuple[int, int]], torch.Tensor]:
    """Load all data in memory. Return a dictionary of indices and a tensor of data."""
    metadata = labels[["#file", "onset", "offset"]].with_row_index()
    lazy = metadata.lazy().sort("#file", maintain_order=True).with_columns(*item_frontiers(frequency))
    indices_lazy = lazy.select("left", "right", "index").sort("index").select("left", "right")
    by_file_lazy = lazy.select("#file", "start", "end").group_by("#file", maintain_order=True).agg("start", "end")
    indices, by_file = pl.collect_all([indices_lazy, by_file_lazy])

    data, device = [], torch.device("cuda" if torch.cuda.is_available() else "cpu")
    for fileid, start_indices, end_indices in tqdm(by_file.iter_rows(), desc="Building dataset", total=len(by_file)):
        features = feature_maker(mapping[fileid]).detach().to(device)
        for start, end in zip(start_indices, end_indices, strict=True):
            if start < 0 or end > features.size(0):
                raise FeaturesSizeError(fileid, start, end, features.size(0))
            if end <= start:
                raise EmptyFeaturesError(
                    lazy.filter(pl.col("end") <= pl.col("start"))
                    .sort("index")
                    .select("#file", "onset", "offset")
                    .collect()
                )
            data.append(features[start:end])
    return dict(enumerate(indices.rows())), torch.cat(data, dim=0)


class TimesArrayDimensionError(ValueError):
    """To raise if the times array is not 1D."""

    def __init__(self) -> None:
        super().__init__("Only 1D times array are supported")


class TimesArrayFrontiersError(ValueError):
    """To raise if we select nothing."""

    def __init__(self, fileid: str, onset: float, offset: float) -> None:
        super().__init__(f"No times were found between onset={onset}, offset={offset} for file {fileid}")


def load_data_from_item_with_times(
    paths_features: dict[str, Path],
    paths_times: dict[str, Path],
    labels: pl.DataFrame,
) -> tuple[dict[int, tuple[int, int]], torch.Tensor]:
    """Load all data in memory using features and times array. This is smaller than using a predefined frequency."""
    metadata = labels[["#file", "onset", "offset"]].with_row_index()
    by_file = (
        metadata.sort("#file", maintain_order=True)
        .group_by("#file", maintain_order=True)
        .agg("index", "onset", "offset")
    )
    data, device, all_indices, right = [], torch.device("cuda" if torch.cuda.is_available() else "cpu"), {}, 0
    decimals = by_file["onset"].dtype.inner.scale  # ty: ignore[unresolved-attribute]
    for fileid, indices, onsets, offsets in tqdm(by_file.iter_rows(), desc="Building dataset", total=len(by_file)):
        features = torch.load(paths_features[fileid], map_location=device).detach()
        times = torch.load(paths_times[fileid]).round(decimals=decimals)
        if times.ndim > 1:
            raise TimesArrayDimensionError
        for index, onset, offset in zip(indices, onsets, offsets, strict=True):
            mask = torch.where(torch.logical_and(float(onset) <= times, times <= float(offset)))[0]  # ty: ignore[invalid-argument-type]
            if not mask.any():
                raise TimesArrayFrontiersError(fileid, float(onset), float(offset))
            data.append(features[mask])
            left = right
            right += len(mask)
            all_indices[index] = (left, right)
    return all_indices, torch.cat(data, dim=0)


@dataclass(frozen=True)
class Dataset:
    """Simple interface to a dataset.

    :param labels: ``pl.DataFrame`` containing the labels of the datapoints.
    :param accessor: ``InMemoryAccessor`` to access the data.
    """

    labels: pl.DataFrame
    accessor: InMemoryAccessor

    def __repr__(self) -> str:
        return f"labels:\n{self.labels!r}\naccessor: {self.accessor!r}"

    def normalize_(self) -> Self:
        """L2 normalization of the data."""
        self.accessor.data = normalize_with_singularity(self.accessor.data.cpu()).to(self.accessor.device)
        return self

    @classmethod
    def from_item(
        cls,
        item: str | Path,
        root: str | Path,
        frequency: int,
        *,
        feature_maker: Callable[[str | Path], torch.Tensor] = torch.load,
        extension: str = ".pt",
    ) -> "Dataset":
        """Create a dataset from an item file.

        If you want to keep the Libri-Light bug to reproduce previous results,
        set the environment variable FASTABX_WITH_LIBRILIGHT_BUG=1.
        """
        labels = read_item(item)
        paths = find_all_files(root, extension)
        indices, data = load_data_from_item(paths, labels, frequency, feature_maker)
        return Dataset(labels=labels, accessor=InMemoryAccessor(indices, data))

    @classmethod
    def from_item_with_times(cls, item: str | Path, features: str | Path, times: str | Path) -> "Dataset":
        """Create a dataset from an item file.

        Use arrays containing the times associated to the features instead of a given frequency.
        """
        labels = read_item(item)
        paths_features = find_all_files(features, ".pt")
        paths_times = find_all_files(times, ".pt")
        indices, data = load_data_from_item_with_times(paths_features, paths_times, labels)
        return Dataset(labels=labels, accessor=InMemoryAccessor(indices, data))

    @classmethod
    def from_item_and_units(
        cls,
        item: str | Path,
        units: str | Path,
        frequency: int,
        *,
        audio_key: str = "audio",
        units_key: str = "units",
        separator: str = " ",
    ) -> "Dataset":
        """Create a dataset from an item file with the units all described in a single JSONL file.

        :param item: Path to the item file.
        :param units: Path to the JSONL file containing the units.
        :param frequency: Frequency of the features.
        :param audio_key: Key in the JSONL file that contains the audio file names.
        :param units_key: Key in the JSONL file that contains the units.
        :param separator: Separator used in the units field.
        """
        labels = read_item(item)
        units_df = (
            pl.scan_ndjson(units)
            .with_columns(pl.col(audio_key).str.split("/").list.last().str.replace(r"\.[^.]+$", ""))
            .collect()
        )

        def feature_maker(idx: int) -> torch.Tensor:
            return torch.tensor([int(unit) for unit in units_df[idx, units_key].split(separator)]).unsqueeze(1)

        mapping: dict[str, int] = dict(zip(units_df[audio_key], range(len(units_df)), strict=True))
        indices, data = load_data_from_item(mapping, labels, frequency, feature_maker)
        return Dataset(labels=labels, accessor=InMemoryAccessor(indices, data))

    @classmethod
    def from_dataframe(cls, df: SupportsInterchange, feature_columns: str | Collection[str]) -> "Dataset":
        """Create a dataset from a DataFrame (polars or pandas)."""
        df = pl.from_dataframe(df.__dataframe__())
        labels = df.select(cs.exclude(feature_columns))
        indices = {i: (i, i + 1) for i in range(len(labels))}
        data = df.select(feature_columns).cast(pl.Float32).to_torch()
        return Dataset(labels=labels, accessor=InMemoryAccessor(indices, data))

    @classmethod
    def from_csv(cls, path: str | Path, feature_columns: str | Collection[str], *, separator: str = ",") -> "Dataset":
        """Create a dataset from a CSV file."""
        return cls.from_dataframe(pl.read_csv(path, separator=separator), feature_columns)

    @classmethod
    def from_dict(cls, data: Mapping[str, Sequence[object]], feature_columns: str | Collection[str]) -> "Dataset":
        """Create a dataset from a dictionary of sequences."""
        return cls.from_dataframe(pl.from_dict(data), feature_columns)

    @classmethod
    def from_dicts(cls, data: Iterable[dict[str, Any]], feature_columns: str | Collection[str]) -> "Dataset":
        """Create a dataset from a sequence of dictionaries."""
        return cls.from_dataframe(pl.from_dicts(data), feature_columns)

    @classmethod
    def from_numpy(
        cls,
        features: ArrayLike,
        labels: Mapping[str, Sequence[object]] | SupportsInterchange,
    ) -> "Dataset":
        """Create a dataset from the features (numpy array) and the labels (dictionary of sequences)."""
        features_df = pl.from_numpy(np.asarray(features))
        labels_df = (
            pl.from_dataframe(labels.__dataframe__()) if hasattr(labels, "__dataframe__") else pl.from_dict(labels)
        )
        if len(features_df) != len(labels_df):
            raise ValueError
        return cls.from_dataframe(pl.concat((features_df, labels_df), how="horizontal"), features_df.columns)


def dummy_dataset_from_item(item: str | Path, frequency: int | None) -> Dataset:
    """To debug."""
    labels = read_item(item).with_columns(pl.lit(0).alias("dummy"))
    if frequency is not None:
        labels = labels.with_columns(*item_frontiers(frequency))
    return Dataset.from_dataframe(labels, "dummy")



"""Task module. The Task class builds all the cells for the 'by', 'on' and 'across' conditions."""

from collections.abc import Generator

from fastabx.cell import Cell, cell_description, cell_header, cells_on_by, cells_on_by_across
from fastabx.dataset import Dataset
from fastabx.subsample import Subsampler
from fastabx.verify import verify_dataset_labels, verify_task_conditions


class Task:
    """The ABX task class.

    A Task builds all the :py:class:`.Cell` given ``on``, ``by`` and ``across`` conditions.
    It can be subsampled to limit the number of cells.
    """

    def __init__(
        self,
        dataset: Dataset,
        *,
        on: str,
        by: list[str] | None = None,
        across: list[str] | None = None,
        subsampler: Subsampler | None = None,
    ) -> None:
        self.dataset = dataset
        self.on = on
        self.by = by or []
        self.across = across or []
        verify_task_conditions([self.on, *self.by, *self.across])
        verify_dataset_labels(dataset.labels.select([self.on, *self.by, *self.across]))
        self._subsampler_description = subsampler.description(with_across=bool(self.across)) if subsampler else ""

        if self.across:
            cells = cells_on_by_across(self.dataset.labels.lazy(), self.on, self.by, self.across)
        else:
            cells = cells_on_by(self.dataset.labels.lazy(), self.on, self.by)
        if subsampler:
            cells = subsampler(cells, with_across=bool(self.across))
        self.cells = cells.with_columns(
            description=cell_description(self.on, self.by, self.across),
            header=cell_header(self.on, self.by, self.across),
        ).collect()

    def __len__(self) -> int:
        return len(self.cells)

    def __getitem__(self, i: int) -> Cell:
        if i < 0 or i >= len(self):
            raise IndexError
        a = self.dataset.accessor.batched(self.cells[i, "index_a"])
        b = self.dataset.accessor.batched(self.cells[i, "index_b"])
        x = self.dataset.accessor.batched(self.cells[i, "index_x"])
        header, description = self.cells[i, "header"], self.cells[i, "description"]
        is_symmetric = not bool(self.across)
        return Cell(a=a, b=b, x=x, header=header, description=description, is_symmetric=is_symmetric)

    def __iter__(self) -> Generator[Cell, None, None]:
        is_symmetric = not bool(self.across)
        columns = ["header", "description", "index_a", "index_b", "index_x"]
        for header, description, index_a, index_b, index_x in self.cells[columns].iter_rows():
            a = self.dataset.accessor.batched(index_a)
            b = self.dataset.accessor.batched(index_b)
            x = self.dataset.accessor.batched(index_x)
            yield Cell(a=a, b=b, x=x, header=header, description=description, is_symmetric=is_symmetric)

    def __repr__(self) -> str:
        return (
            f"Task(\n\tON({self.on})"
            + (f"\n\tBY({', '.join(self.by)})" if self.by else "")
            + (f"\n\tACROSS({', '.join(self.across)})" if self.across else "")
            + f"\n\t{self._subsampler_description}\n)"
        )



"""Various utilities."""

import importlib.resources
import json
import os
import sys

import torch

MIN_CELLS_FOR_TQDM = 50


class PyTorchVersionError(ImportError):
    """PyTorch version mismatch."""

    def __init__(self, actual: str, expected: str) -> None:
        super().__init__(
            f"The DTW extension requires PyTorch {expected}, but you have {actual}. "
            "Please install the correct version of PyTorch, or install the wheel of fastabx "
            "from the GitHub release page that matches your PyTorch version."
        )


def load_dtw_extension() -> None:
    """Load the DTW extension.

    We check that PyTorch has been installed with the correct version.
    """
    expected = json.loads((importlib.resources.files("fastabx") / "torch_version.json").read_text())[sys.platform]
    if torch.__version__ != expected:
        raise PyTorchVersionError(torch.__version__, expected)
    from . import _C  # ty: ignore[unresolved-import] # noqa: F401, PLC0415


def with_librilight_bug() -> bool:
    """Whether to reproduce the results from LibriLight ABX or not."""
    return os.getenv("FASTABX_WITH_LIBRILIGHT_BUG", "0") == "1"



"""Compare CPU and CUDA dtw implementations."""

import pytest
import torch
from hypothesis import given, settings
from hypothesis import strategies as st

from fastabx.dtw import dtw, dtw_batch

rtol, atol = 0, 1e-9
skipifnogpu = pytest.mark.skipif(not torch.cuda.is_available(), reason="No GPU available")

DIM, BATCH = st.integers(1, 1280), st.integers(1, 3)
LOW, HIGH_MINUS_LOW = st.floats(-100, 100), st.floats(0.1, 100)


def make_tensor(shape: tuple[int, ...], *, dtype: torch.dtype, low: float, high: float) -> torch.Tensor:
    """Build a tensor for testing."""
    if low == high and dtype == torch.long:
        return torch.ones(shape, dtype=torch.long, device="cpu")
    return torch.testing.make_tensor(shape, dtype=dtype, device="cpu", low=low, high=high)


@skipifnogpu
@given(x=DIM, y=DIM, low=LOW, high_minus_low=HIGH_MINUS_LOW)
@settings(deadline=None)
def test_dtw(x: int, y: int, low: float, high_minus_low: float) -> None:
    """Compare the output of dtw between CPU and GPU implementations."""
    d = make_tensor((x, y), dtype=torch.float32, low=low, high=high_minus_low + low)
    torch.testing.assert_close(dtw(d), dtw(d.cuda()).cpu(), rtol=rtol, atol=atol)


@skipifnogpu
@given(n=BATCH, x=DIM, low=LOW, high_minus_low=HIGH_MINUS_LOW)
@settings(deadline=None)
def test_dtw_batch_symmetric(n: int, x: int, low: float, high_minus_low: float) -> None:
    """Compare the output of dtw_batch between CPU and GPU implementations, symmetric case."""
    d = make_tensor((n, n, x, x), dtype=torch.float32, low=low, high=high_minus_low + low)
    sx = make_tensor((n,), dtype=torch.long, low=1, high=x)
    i, j = torch.triu_indices(n, n)
    d[i, j] = d[j, i]
    torch.testing.assert_close(
        dtw_batch(d, sx, sx, symmetric=True),
        dtw_batch(d.cuda(), sx.cuda(), sx.cuda(), symmetric=True).cpu(),
        rtol=rtol,
        atol=atol,
    )


@skipifnogpu
@given(n=BATCH, m=BATCH, x=DIM, y=DIM, low=LOW, high_minus_low=HIGH_MINUS_LOW)
@settings(deadline=None)
def test_dtw_batch_not_symmetric(n: int, m: int, x: int, y: int, low: float, high_minus_low: float) -> None:
    """Compare the output of dtw_batch between CPU and GPU implementations, non symmetric case."""
    d = make_tensor((n, m, x, y), dtype=torch.float32, low=low, high=high_minus_low + low)
    sx = make_tensor((n,), dtype=torch.long, low=1, high=x)
    sy = make_tensor((m,), dtype=torch.long, low=1, high=y)
    torch.testing.assert_close(
        dtw_batch(d, sx, sy, symmetric=False),
        dtw_batch(d.cuda(), sx.cuda(), sy.cuda(), symmetric=False).cpu(),
        rtol=rtol,
        atol=atol,
    )



"""Safety checks for the public API."""

import enum
from collections.abc import Sequence
from itertools import chain

import polars as pl
from torch import Tensor

NDIM = 3
INVALID_COLUMN_SUFFIX = ("_a", "_b", "_x")
INVALID_COLUMN_NAMES = {"index", "score", "size", "is_valid", "__group", "__lookup"}


class LabelReservedNameError(ValueError):
    """Invalid name for a condition."""

    def __init__(self, name: str) -> None:
        super().__init__(f"Invalid label: {name}. This name is reserved for internal computations.")


class LabelSuffixError(ValueError):
    """Invalid suffix for a condition."""

    def __init__(self, name: str) -> None:
        super().__init__(f"Invalid label: {name}. Cannot end by _a, _b, or _x.")


class EmptyDataPointsError(ValueError):
    """Empty data points in the dataset."""

    max_print_size: int = 10

    def __init__(self, empty: list[str]) -> None:
        super().__init__(
            f"{len(empty)} empty elements were found in the dataset (with indices ["
            + ", ".join(empty[: self.max_print_size])
            + (", ..." if len(empty) > self.max_print_size else "")
            + "])"
        )


class DuplicateConditionsError(ValueError):
    """Duplicate conditions found."""


class InputTypeError(TypeError):
    """All conditions should be strings."""

    def __init__(self, expected: type, received: type) -> None:
        super().__init__(f"Should be an instance of {expected}, not {received}")


def verify_empty_datapoints(indices: dict[int, tuple[int, int]]) -> None:
    """Check if some datapoints are empty by checking the start and end indices."""
    empty = []
    for key, (start, end) in indices.items():
        if end <= start:
            empty.append(str(key))
    if empty:
        raise EmptyDataPointsError(empty)


def verify_task_conditions(conditions: list[str]) -> None:
    """Conditions should be unique strings."""
    if len(conditions) != len(set(conditions)):
        raise DuplicateConditionsError
    for cond in conditions:
        if not isinstance(cond, str):
            raise InputTypeError(str, type(cond))


def verify_dataset_labels(df: pl.DataFrame) -> None:
    """Check the column labels."""
    for col in df.schema:
        if col in INVALID_COLUMN_NAMES:
            raise LabelReservedNameError(col)
        if col.endswith(INVALID_COLUMN_SUFFIX):
            raise LabelSuffixError(col)


def verify_subsampler_params(*sizes: int | None, seed: int) -> None:
    """All sizes must be positive integers."""
    if not all(isinstance(s, int) and s > 1 for s in sizes if s is not None):
        msg = "sizes should be positive integers"
        raise TypeError(msg)
    if not isinstance(seed, int):
        raise InputTypeError(int, type(seed))


class CellErrorType(enum.Enum):
    """All types of errors coming from a ``Cell``."""

    NDIM = enum.auto()
    FEATURE_DIM = enum.auto()
    SIZE = enum.auto()


class InvalidCellError(ValueError):
    """The cell is not built correctly."""

    def __init__(self, error_type: CellErrorType) -> None:
        match error_type:
            case CellErrorType.NDIM:
                msg = "A, B, and X should be tensors with 3 dimensions"
            case CellErrorType.FEATURE_DIM:
                msg = "A, B, and X should have the same feature dimension"
            case CellErrorType.SIZE:
                msg = "Invalid size specification"
            case _:
                msg = None
        super().__init__(msg)


def verify_cell(a_sa: tuple[Tensor, Tensor], b_sb: tuple[Tensor, Tensor], x_sx: tuple[Tensor, Tensor]) -> None:
    """Assert the integrity of a cell."""
    (a, sa), (b, sb), (x, sx) = a_sa, b_sb, x_sx
    if not a.ndim == b.ndim == x.ndim == NDIM:
        raise InvalidCellError(CellErrorType.NDIM)
    if not a.size(2) == b.size(2) == x.size(2):
        raise InvalidCellError(CellErrorType.FEATURE_DIM)
    if not (a.size(0) == sa.size(0) and b.size(0) == sb.size(0) and x.size(0) == sx.size(0)):
        raise InvalidCellError(CellErrorType.SIZE)


class LevelsErrorType(enum.Enum):
    """All types of errors coming that can arise from 'levels'."""

    FORMAT = enum.auto()
    DUPLICATES = enum.auto()
    COLUMNS = enum.auto()


class InvalidLevelsError(ValueError):
    """Levels are not well formatted."""

    def __init__(self, error_type: LevelsErrorType) -> None:
        match error_type:
            case LevelsErrorType.FORMAT:
                msg = "'levels' should be list[tuple[str, ...] | str]"
            case LevelsErrorType.DUPLICATES:
                msg = "levels should not contain duplicates"
            case LevelsErrorType.COLUMNS:
                msg = "levels should be columns of the DataFrame"
            case _:
                msg = None
        super().__init__(msg)


def format_score_levels(levels: Sequence[tuple[str, ...] | str]) -> list[tuple[str, ...]]:
    """Put all the levels in tuples."""
    formatted: list[tuple[str, ...]] = []
    for level in levels:
        if isinstance(level, str):
            formatted.append((level,))
        elif isinstance(level, tuple) and all(isinstance(x, str) for x in level):
            formatted.append(level)
        else:
            raise InvalidLevelsError(LevelsErrorType.FORMAT)
    return formatted


def verify_score_levels(columns: list[str], levels: list[tuple[str, ...]]) -> None:
    """Levels should be unique columns of the DataFrame."""
    all_levels = list(chain.from_iterable(levels))
    unique_levels = set(all_levels)
    if len(all_levels) != len(unique_levels):
        raise InvalidLevelsError(LevelsErrorType.DUPLICATES)
    if not unique_levels.issubset(set(columns)):
        raise InvalidLevelsError(LevelsErrorType.COLUMNS)



"""DTW implementation using PyTorch C++ extensions, with CPU and CUDA backends."""

import torch


def dtw(distances: torch.Tensor) -> torch.Tensor:
    """Compute the DTW of the given ``distances`` 2D tensor."""
    return torch.ops.fastabx.dtw.default(distances)


def dtw_batch(distances: torch.Tensor, sx: torch.Tensor, sy: torch.Tensor, *, symmetric: bool) -> torch.Tensor:
    """Compute the batched DTW on the ``distances`` 4D tensor."""
    return torch.ops.fastabx.dtw_batch.default(distances, sx, sy, symmetric)


@torch.library.register_fake("fastabx::dtw")
def _(distances: torch.Tensor) -> torch.Tensor:
    """Register the FakeTensor kernel for dtw, for compatibility with torch.compile."""
    torch._check(distances.ndim == 2)
    torch._check(distances.dtype == torch.float32)
    return torch.empty((), dtype=torch.float32, layout=distances.layout, device=distances.device)


@torch.library.register_fake("fastabx::dtw_batch")
def _(distances: torch.Tensor, sx: torch.Tensor, sy: torch.Tensor, symmetric: bool) -> torch.Tensor:  # noqa: FBT001
    """Register the FakeTensor kernel for dtw_batch, for compatibility with torch.compile."""
    torch._check(distances.ndim == 4)
    torch._check(sx.ndim == 1)
    torch._check(sy.ndim == 1)
    torch._check(distances.dtype == torch.float32)
    torch._check(sx.dtype == torch.long)
    torch._check(sy.dtype == torch.long)
    torch._check(isinstance(symmetric, bool))
    nx, ny, _, _ = distances.shape
    return torch.empty((nx, ny), dtype=torch.float32, layout=distances.layout, device=distances.device)



"""Score the ABX task for each cell and collapse the scores into a final score."""

from collections.abc import Sequence
from pathlib import Path

import polars as pl
import polars.selectors as cs
from tqdm import tqdm

from fastabx.constraints import Constraints, score_task_with_constraints
from fastabx.distance import Distance, DistanceName, abx_on_cell, distance_function
from fastabx.task import Task
from fastabx.utils import MIN_CELLS_FOR_TQDM
from fastabx.verify import format_score_levels, verify_score_levels


def pl_weighted_mean(value_col: str, weight_col: str) -> pl.Expr:
    """Generate a Polars aggregation expression to take a weighted mean.

    https://github.com/pola-rs/polars/issues/7499#issuecomment-2569748864
    """
    values = pl.col(value_col)
    weights = pl.when(values.is_not_null()).then(weight_col)
    return weights.dot(values).truediv(weights.sum()).fill_nan(None)


class CollapseError(Exception):
    """Something wrong happened when collapsing the ``Score``."""

    def __init__(self, *, are_set: bool) -> None:
        if are_set:
            msg = "Cannot set `weighted=True` and `levels` at the same time."
        else:
            msg = "Either set `levels` or `weighted=True`."
        super().__init__(msg)


def score_details(cells: pl.DataFrame, *, levels: Sequence[tuple[str, ...] | str] | None) -> pl.DataFrame:
    """Collapse the scored cells and return the final scores and sizes for each (A, B) pairs."""
    if levels is None:
        if len(set(cells.columns) - {"index_a", "index_b", "index_x", "score", "size"}) != 2:
            raise CollapseError(are_set=False)
        levels = []
    cells = cells.select(~(cs.starts_with("index") | cs.ends_with("_x")))
    levels_in_tuples = format_score_levels(levels)
    verify_score_levels(cells.columns, levels_in_tuples)
    for level in levels_in_tuples:
        group_key = cs.exclude("score", "size", *level)
        cells = cells.group_by(group_key, maintain_order=True).agg(pl.col("score").mean(), pl.col("size").sum())
    return cells


def score_task(task: Task, distance: Distance) -> tuple[list[float], list[int]]:
    """Score each cell of a :py:class:`.Task` using a given distance, and return scores and sizes."""
    scores, sizes = [], []
    for cell in tqdm(task, "Scoring each cell", disable=len(task) < MIN_CELLS_FOR_TQDM):
        scores.append(abx_on_cell(cell, distance).item())
        sizes.append(len(cell))
    return scores, sizes


class Score:
    """Compute the score of a :py:class:`.Task` using a given distance specified by ``distance_name``.

    Additional :py:type:`.Constraints` can be provided to restrict the possible triplets in each cell.
    """

    def __init__(self, task: Task, distance_name: DistanceName, *, constraints: Constraints | None = None) -> None:
        self.distance_name = distance_name
        distance = distance_function(distance_name)
        if distance_name in {"cosine", "angular"}:
            task.dataset.normalize_()
        scores, sizes = (
            score_task(task, distance)
            if constraints is None
            else score_task_with_constraints(task, distance, constraints)
        )
        self._cells = task.cells.select(cs.exclude("description", "header")).with_columns(
            score=pl.Series(scores, dtype=pl.Float32), size=pl.Series(sizes)
        )

    @property
    def cells(self) -> pl.DataFrame:
        """Return the scored cells."""
        return self._cells

    @cells.setter
    def cells(self, _: pl.DataFrame) -> None:
        msg = "The `cells` attribute is read-only."
        raise AttributeError(msg)

    def __repr__(self) -> str:
        return f"Score({len(self.cells)} cells, {self.distance_name} distance)"

    def write_csv(self, file: str | Path) -> None:
        """Write the results of all the cells to a CSV file."""
        nested = [name for name, dtype in self.cells.schema.items() if dtype == pl.List]
        (self.cells.select(cs.exclude(nested)) if nested else self.cells).write_csv(file)

    def details(self, *, levels: Sequence[tuple[str, ...] | str] | None) -> pl.DataFrame:
        """Collapse the scored cells and return the final scores and sizes for each (A, B) pairs.

        :param levels: List of levels to collapse. The order matters a lot.
        """
        return score_details(self.cells, levels=levels)

    def collapse(self, *, levels: Sequence[tuple[str, ...] | str] | None = None, weighted: bool = False) -> float:
        """Collapse the scored cells into the final score.

        Use either `levels` or `weighted=True` to collapse the scores.

        :param levels: List of levels to collapse. The order matters a lot.
        :param weighted: Whether to collapse the scores using a mean weighted by the size of the cells.
        """
        if weighted:
            if levels is not None:
                raise CollapseError(are_set=True)
            return self.cells.select(pl_weighted_mean("score", "size")).item()
        return self.details(levels=levels)["score"].mean()



"""Verify that we reproduce the ABX error rate from the Zerospeech library using features from HuBERT base L11."""

from pathlib import Path
from typing import Literal

import pytest
import torch

from fastabx import zerospeech_abx


@pytest.fixture
def item(request: pytest.FixtureRequest) -> Path:
    """Item file."""
    path = Path(request.config.getoption("--item"))
    if not path.is_file():
        pytest.fail(f"Item file not found: {path}")
    if path.name not in request.config.reference_scores:
        pytest.fail(f"Invalid item, must be one of {set(request.config.reference_scores)})")
    return path


@pytest.fixture
def features(request: pytest.FixtureRequest) -> Path:
    """Features directory."""
    path = Path(request.config.getoption("--features"))
    if not path.is_dir():
        pytest.fail(f"Features directory not found: {path}")
    return path


@pytest.mark.skipif("not config.getoption('item') or not config.getoption('features')")
@pytest.mark.parametrize("speaker", ["within", "across"])
@pytest.mark.parametrize("context", ["within", "any"])
def test_zerospeech(
    pytestconfig: pytest.Config,
    item: Path,
    features: Path,
    speaker: Literal["within", "across"],
    context: Literal["within", "any"],
) -> None:
    """Test reproducibility."""
    if (speaker, context) not in pytestconfig.reference_scores[item.name]:
        pytest.skip(f"Configuration not supported for {item.stem}: {speaker} speaker, {context} context")
    reference = pytestconfig.reference_scores[item.name][(speaker, context)]
    score = zerospeech_abx(
        item,
        features,
        max_size_group=pytestconfig.max_size_group,
        max_x_across=pytestconfig.max_x_across,
        speaker=speaker,
        context=context,
        distance=pytestconfig.distance,
        frequency=pytestconfig.frequency,
        seed=pytestconfig.seed,
    )
    torch.testing.assert_close(score, reference, rtol=0, atol=1e-5)



"""Build the DTW PyTorch C++ extension."""

import os
import sys

from setuptools import Extension, setup
from torch.utils.cpp_extension import CUDA_HOME, BuildExtension, CppExtension, CUDAExtension


def get_openmp_flags() -> tuple[list[str], list[str]]:
    """Return the compiler and linker flags for OpenMP."""
    match sys.platform:
        case "linux":
            compile_flags, link_flags = ["-fopenmp"], ["-fopenmp"]
        case "win32":
            compile_flags, link_flags = ["-openmp"], []
        case _:  # On MacOS, we use the OpenMP version vendored by PyTorch
            return [], []
    return compile_flags, link_flags


def get_extension() -> Extension:
    """Either CUDA or CPU extension."""
    use_cuda = CUDA_HOME is not None
    extension = CUDAExtension if use_cuda else CppExtension
    openmp_flags = get_openmp_flags()
    extra_compile_args = {
        "cxx": ["-fdiagnostics-color=always", "-DPy_LIMITED_API=0x030C0000", "-O3"] + openmp_flags[0],
        "nvcc": ["-O3"],
    }
    sources = ["src/fastabx/csrc/dtw.cpp"]
    if use_cuda:
        os.environ["TORCH_CUDA_ARCH_LIST"] = "Volta;Turing;Ampere;Ada;Hopper"
        sources.append("src/fastabx/csrc/cuda/dtw.cu")
    return extension(
        "fastabx._C",
        sources,
        extra_compile_args=extra_compile_args,
        extra_link_args=openmp_flags[1],
        py_limited_api=True,
    )


if __name__ == "__main__":
    setup(
        ext_modules=[get_extension()],
        cmdclass={"build_ext": BuildExtension},
        options={"bdist_wheel": {"py_limited_api": "cp312"}},
    )



"""Pooling utilities."""

from collections.abc import Callable
from dataclasses import dataclass
from functools import partial
from typing import Literal

import torch

from fastabx.dataset import Dataset, InMemoryAccessor

type PoolingName = Literal["mean", "hamming"]


def hamming_window(x: torch.Tensor) -> torch.Tensor:
    """Apply the hamming window on the input Tensor."""
    window = torch.hamming_window(x.size(0), device=x.device)
    return (window @ x) / window.sum()


def pooling_function(name: PoolingName) -> Callable[[torch.Tensor], torch.Tensor]:
    """Return the corresponding pooling function."""
    match name:
        case "mean":
            return partial(torch.mean, dim=0)
        case "hamming":
            return hamming_window
        case _:
            raise ValueError(name)


@dataclass(frozen=True)
class PooledDataset(Dataset):
    """Pooled dataset."""

    pooling: PoolingName

    def __repr__(self) -> str:
        return f"labels:\n{self.labels!r}\naccessor: {self.accessor!r}\npooling: {self.pooling}"


def pooling(dataset: Dataset, pooling_name: PoolingName) -> PooledDataset:
    """Pool the :py:class:`.Dataset` using the pooling method given by ``pooling_name``.

    The pooled dataset is a new one, with data stored in memory.
    For simplicity, we iterate through the original dataset and
    apply pooling on each element.
    """
    labels = dataset.labels
    indices = {i: (i, i + 1) for i in range(len(labels))}
    pooling_fn = pooling_function(pooling_name)
    data = torch.stack([pooling_fn(x) for x in dataset.accessor], dim=0)
    return PooledDataset(pooling=pooling_name, labels=labels, accessor=InMemoryAccessor(indices, data))



"""Subsampling functions."""

import polars as pl
import polars.selectors as cs

from fastabx.verify import verify_subsampler_params


def subsample_each_cell(df: pl.LazyFrame, size: int, seed: int) -> pl.LazyFrame:
    """Subsample each cell by taking at most ``size`` instances of A, B, and X independently."""
    return (
        df.with_columns(pl.concat_str(~cs.starts_with("index"), separator="-").alias("__group"))
        .with_columns(cs.starts_with("index").explode().shuffle(seed=seed).implode().over("__group").list.head(size))
        .select(cs.exclude("__group"))
    )


def subsample_across_group(df: pl.LazyFrame, size: int, seed: int) -> pl.LazyFrame:
    """Subsample each group of 'across' condition by taking ``size`` possible values for X in each group."""
    x_cols = [c for c in df.collect_schema() if c.endswith("_x") and c != "index_x"]
    to_ignore = cs.starts_with("index") | cs.ends_with("_x")
    df = df.with_columns(pl.concat_str(~to_ignore, separator="-").alias("__group"))
    return (
        df.group_by("__group", maintain_order=True)
        .agg((cs.ends_with("_x") & (~cs.starts_with("index"))).unique(maintain_order=True).shuffle(seed).head(size))
        .explode(x_cols)
        .join(df, on=["__group", *x_cols], how="left")
        .select(cs.exclude("__group"))
    )


class Subsampler:
    """Subsample the ABX :py:class:`.Task`.

    Each cell is limited to ``max_size_group`` items for A, B and X independently.
    When using "across" conditions, each group of (A, B) is limited to ``max_x_across`` possible values for X.
    Subsampling for one or more conditions can be disabled by setting the corresponding argument to ``None``.
    """

    def __init__(self, max_size_group: int | None, max_x_across: int | None, seed: int = 0) -> None:
        verify_subsampler_params(max_size_group, max_x_across, seed=seed)
        self.max_size_group = max_size_group
        self.max_x_across = max_x_across
        self.seed = seed

    def __call__(self, lazy_cells: pl.LazyFrame, *, with_across: bool) -> pl.LazyFrame:
        """Subsample the cells."""
        if self.max_size_group is not None:
            lazy_cells = subsample_each_cell(lazy_cells, self.max_size_group, self.seed)
        if with_across and self.max_x_across is not None:
            lazy_cells = subsample_across_group(lazy_cells, self.max_x_across, self.seed)
        return lazy_cells

    def description(self, *, with_across: bool) -> str:
        """Return a description of the subsampling."""
        desc = []
        if self.max_size_group is not None:
            desc.append(f"maximal number of A, B, or X in a cell: {self.max_size_group}")
        if with_across and self.max_x_across is not None:
            desc.append(f"maximal number of X for (A, B): {self.max_x_across}")
        return ",".join(desc)



"""Check for compatibility with torch.compile."""

import torch
from hypothesis import given, settings
from hypothesis import strategies as st
from torch.library import opcheck

import fastabx  # noqa: F401 # Need to import it to register dtw operation

DIM, BATCH = st.integers(1, 1280), st.integers(1, 3)
LOW, HIGH_MINUS_LOW = st.floats(-100, 100), st.floats(0.1, 100)
CUDA_AVAILABLE = torch.cuda.is_available()


def make_tensor(shape: tuple[int, ...], *, dtype: torch.dtype, low: float, high: float) -> torch.Tensor:
    """Build a tensor for testing."""
    if low == high and dtype == torch.long:
        return torch.ones(shape, dtype=torch.long, device="cpu")
    return torch.testing.make_tensor(shape, dtype=dtype, device="cpu", low=low, high=high)


@given(x=DIM, y=DIM, low=LOW, high_minus_low=HIGH_MINUS_LOW)
@settings(deadline=None)
def test_opcheck_dtw(x: int, y: int, low: float, high_minus_low: float) -> None:
    """Verify that dtw can be torch compiled."""
    sample = make_tensor((x, y), dtype=torch.float32, low=low, high=high_minus_low + low)
    opcheck(torch.ops.fastabx.dtw.default, (sample,))
    if CUDA_AVAILABLE:
        opcheck(torch.ops.fastabx.dtw.default, (sample.cuda(),))


@given(n=BATCH, x=DIM, low=LOW, high_minus_low=HIGH_MINUS_LOW)
@settings(deadline=None)
def test_opcheck_dtw_batch_symmetric(n: int, x: int, low: float, high_minus_low: float) -> None:
    """Verify that dtw_batch can be torch compiled, with symmetric input."""
    sample = make_tensor((n, n, x, x), dtype=torch.float32, low=low, high=high_minus_low + low)
    sx = make_tensor((n,), dtype=torch.long, low=1, high=x)
    i, j = torch.triu_indices(n, n)
    sample[i, j] = sample[j, i]
    opcheck(torch.ops.fastabx.dtw_batch.default, (sample, sx, sx), {"symmetric": True})
    if CUDA_AVAILABLE:
        opcheck(torch.ops.fastabx.dtw_batch.default, (sample.cuda(), sx.cuda(), sx.cuda()), {"symmetric": True})


@given(n=BATCH, m=BATCH, x=DIM, y=DIM, low=LOW, high_minus_low=HIGH_MINUS_LOW)
@settings(deadline=None)
def test_opcheck_dtw_batch_not_symmetric(n: int, m: int, x: int, y: int, low: float, high_minus_low: float) -> None:
    """Verify that dtw_batch can be torch compiled, with symmetric input."""
    sample = make_tensor((n, m, x, y), dtype=torch.float32, low=low, high=high_minus_low + low)
    sx = make_tensor((n,), dtype=torch.long, low=1, high=x)
    sy = make_tensor((m,), dtype=torch.long, low=1, high=y)
    opcheck(torch.ops.fastabx.dtw_batch.default, (sample, sx, sy), {"symmetric": False})
    if CUDA_AVAILABLE:
        opcheck(torch.ops.fastabx.dtw_batch.default, (sample.cuda(), sx.cuda(), sy.cuda()), {"symmetric": False})
